import pytest
import textwrap
import transaction

from decimal import Decimal
from onegov.directory import DirectoryCollection
from onegov.directory import DirectoryConfiguration
from onegov.org.models import Organisation
from onegov.winterthur.daycare import DaycareSubsidyCalculator, Services


@pytest.fixture(scope='function')
def app(winterthur_app):
    app = winterthur_app

    session = app.session()

    dirs = DirectoryCollection(session, type='extended')
    directory = dirs.add(
        title="Daycare Centers",
        structure=textwrap.dedent("""
            Name *= ___
            Webseite = https://
            Tagestarif *= 0..1000
            Öffnungswochen *= 0..52
        """),
        configuration=DirectoryConfiguration(
            title="[Name]",
            order=('Name', ),
        ))

    # Some actual daycare centers in Winterthur
    directory.add(values=dict(
        name="Pinochio",
        tagestarif=98,
        offnungswochen=49,
        webseite="",
    ))

    directory.add(values=dict(
        name="Fantasia",
        tagestarif=108,
        offnungswochen=51,
        webseite="",
    ))

    directory.add(values=dict(
        name="Kinderhaus",
        tagestarif=110,
        offnungswochen=50,
        webseite="",
    ))

    directory.add(values=dict(
        name="Luftibus",
        tagestarif=110,
        offnungswochen=51,
        webseite="",
    ))

    directory.add(values=dict(
        name="Child Care Corner",
        tagestarif=125,
        offnungswochen=51,
        webseite="",
    ))

    directory.add(values=dict(
        name="Apfelblüte",
        tagestarif=107,
        offnungswochen=51,
        webseite="",
    ))

    directory.add(values=dict(
        name="Am Park",
        tagestarif=120,
        offnungswochen=49,
        webseite="",
    ))

    org = session.query(Organisation).one()
    org.meta['daycare_settings'] = {
        'rebate': Decimal('5.00'),
        'max_rate': Decimal('107'),
        'min_rate': Decimal('15'),
        'max_income': Decimal('75000'),
        'max_wealth': Decimal('154000'),
        'min_income': Decimal('20000'),
        'max_subsidy': Decimal('92'),
        'wealth_premium': Decimal('10.00'),
        'directory': directory.id.hex,
        'services': textwrap.dedent("""
            - titel: "Ganzer Tag inkl. Mitagessen"
              tage: "Montag, Dienstag, Mittwoch, Donnerstag, Freitag"
              prozent: 100.00

            - titel: "Vor- oder Nachmittag inkl. Mitagessen"
              tage: "Montag, Dienstag, Mittwoch, Donnerstag, Freitag"
              prozent: 75.00

            - titel: "Vor- oder Nachmittag ohne Mitagessen"
              tage: "Montag, Dienstag, Mittwoch, Donnerstag, Freitag"
              prozent: 50.00
        """)
    }

    transaction.commit()
    session.close_all()

    return app


def test_calculate_example_1(app):
    calculator = DaycareSubsidyCalculator(app.session())

    services = Services.from_org(app.org)
    services.select('ganzer-tag-inkl-mitagessen', 'mo')
    services.select('ganzer-tag-inkl-mitagessen', 'di')
    services.select('ganzer-tag-inkl-mitagessen', 'mi')
    services.select('ganzer-tag-inkl-mitagessen', 'do')
    services.select('ganzer-tag-inkl-mitagessen', 'fr')

    calculation = calculator.calculate(
        daycare=calculator.daycare_by_title("Fantasia"),
        services=services,
        income=Decimal('75000'),
        wealth=Decimal('150000'),
        rebate=True,
    )

    base, gross, net, actual, monthly = calculation.blocks

    # note, these results slightly differ from the output, as the rounding
    # only happens when the numbers are rendered
    results = [(r.title, r.operation, r.amount) for r in base.results]
    assert results == [
        ("Steuerbares Einkommen", None, Decimal('75000')),
        ("Vermögenszuschlag", "+", Decimal('0')),
        ("Massgebendes Gesamteinkommen", "=", Decimal('75000')),
        ("Abzüglich Minimaleinkommen", "-", Decimal('20000')),
        ("Berechnungsgrundlage", "=", Decimal('55000')),
    ]

    results = [(r.title, r.operation, r.amount) for r in gross.results]
    assert results == [
        ("Übertrag", None, Decimal('55000')),
        ("Faktor", "×", Decimal("0.001672727")),
        ("Einkommensabhängiger Elternbeitragsbestandteil", "=", Decimal("92")),
        ("Mindestbeitrag Eltern", "+", Decimal("15")),
        ("Elternbeitrag brutto", "=", Decimal("107")),
    ]

    results = [(r.title, r.operation, r.amount) for r in net.results]
    assert results == [
        ("Übertrag", None, Decimal('107')),
        ("Rabatt", "-", Decimal('5.35')),
        ("Elternbeitrag netto", "=", Decimal('101.65')),
    ]

    results = [(r.title, r.operation, r.amount) for r in actual.results]
    assert results == [
        ("Übertrag", None, Decimal('101.65')),
        ("Zusatzbeitrag Eltern", "+", Decimal('1')),
        ("Elternbeitrag pro Tag", "=", Decimal('102.65')),
        ("Städtischer Beitrag pro Tag", None, Decimal('5.35'))
    ]

    results = [(r.title, r.operation, r.amount) for r in monthly.results]
    assert results == [
        ("Wochentarif", None, Decimal('513.25')),
        ("Faktor", "×", Decimal('4.25')),
        ("Elternbeitrag pro Monat", "=", Decimal('2181.31')),
        ("Städtischer Beitrag pro Monat", None, Decimal('113.69')),
    ]


def test_calculate_example_2(app):
    calculator = DaycareSubsidyCalculator(app.session())

    services = Services.from_org(app.org)
    services.select('ganzer-tag-inkl-mitagessen', 'mo')
    services.select('ganzer-tag-inkl-mitagessen', 'di')
    services.select('ganzer-tag-inkl-mitagessen', 'mi')
    services.select('ganzer-tag-inkl-mitagessen', 'do')
    services.select('ganzer-tag-inkl-mitagessen', 'fr')

    calculation = calculator.calculate(
        daycare=calculator.daycare_by_title("Apfelblüte"),
        services=services,
        income=Decimal('25000'),
        wealth=Decimal('0'),
        rebate=False,
    )

    base, gross, net, actual, monthly = calculation.blocks

    # note, these results slightly differ from the output, as the rounding
    # only happens when the numbers are rendered
    results = [(r.title, r.amount) for r in base.results]
    assert results == [
        ("Steuerbares Einkommen", Decimal('25000')),
        ("Vermögenszuschlag", Decimal('0')),
        ("Massgebendes Gesamteinkommen", Decimal('25000')),
        ("Abzüglich Minimaleinkommen", Decimal('20000')),
        ("Berechnungsgrundlage", Decimal('5000')),
    ]

    results = [(r.title, r.amount) for r in gross.results]
    assert results == [
        ("Übertrag", Decimal('5000')),
        ("Faktor", Decimal("0.001672727")),
        ("Einkommensabhängiger Elternbeitragsbestandteil", Decimal("8.36")),
        ("Mindestbeitrag Eltern", Decimal("15")),
        ("Elternbeitrag brutto", Decimal("23.36")),
    ]

    results = [(r.title, r.amount) for r in net.results]
    assert results == [
        ("Übertrag", Decimal('23.36')),
        ("Rabatt", Decimal('0')),
        ("Elternbeitrag netto", Decimal('23.36')),
    ]

    results = [(r.title, r.amount) for r in actual.results]
    assert results == [
        ("Übertrag", Decimal('23.36')),
        ("Zusatzbeitrag Eltern", Decimal('0')),
        ("Elternbeitrag pro Tag", Decimal('23.36')),
        ("Städtischer Beitrag pro Tag", Decimal('83.64'))
    ]

    results = [(r.title, r.amount) for r in monthly.results]
    assert results == [
        ("Wochentarif", Decimal('116.80')),
        ("Faktor", Decimal('4.2500')),
        ("Elternbeitrag pro Monat", Decimal('496.40')),
        ("Städtischer Beitrag pro Monat", Decimal('1777.35')),
    ]


def test_calculate_example_3(app):
    calculator = DaycareSubsidyCalculator(app.session())

    services = Services.from_org(app.org)
    services.select('ganzer-tag-inkl-mitagessen', 'mo')

    calculation = calculator.calculate(
        daycare=calculator.daycare_by_title("Pinochio"),
        services=services,
        income=Decimal('61000'),
        wealth=Decimal('264000'),
        rebate=True,
    )

    base, gross, net, actual, monthly = calculation.blocks

    # note, these results slightly differ from the output, as the rounding
    # only happens when the numbers are rendered
    results = [(r.title, r.amount) for r in base.results]
    assert results == [
        ("Steuerbares Einkommen", Decimal('61000')),
        ("Vermögenszuschlag", Decimal('11000')),
        ("Massgebendes Gesamteinkommen", Decimal('72000')),
        ("Abzüglich Minimaleinkommen", Decimal('20000')),
        ("Berechnungsgrundlage", Decimal('52000')),
    ]

    results = [(r.title, r.amount) for r in gross.results]
    assert results == [
        ("Übertrag", Decimal('52000')),
        ("Faktor", Decimal("0.001509091")),
        ("Einkommensabhängiger Elternbeitragsbestandteil", Decimal("78.47")),
        ("Mindestbeitrag Eltern", Decimal("15")),
        ("Elternbeitrag brutto", Decimal("93.47")),
    ]

    results = [(r.title, r.amount) for r in net.results]
    assert results == [
        ("Übertrag", Decimal('93.47')),
        ("Rabatt", Decimal('4.67')),
        ("Elternbeitrag netto", Decimal('88.80')),
    ]

    results = [(r.title, r.amount) for r in actual.results]
    assert results == [
        ("Übertrag", Decimal('88.80')),
        ("Zusatzbeitrag Eltern", Decimal('0')),
        ("Elternbeitrag pro Tag", Decimal('88.80')),
        ("Städtischer Beitrag pro Tag", Decimal('9.20'))
    ]

    results = [(r.title, r.amount) for r in monthly.results]
    assert results == [
        ("Wochentarif", Decimal('88.80')),
        ("Faktor", Decimal('4.0833')),
        ("Elternbeitrag pro Monat", Decimal('362.60')),
        ("Städtischer Beitrag pro Monat", Decimal('37.57')),
    ]


def test_caclulate_example_4(app):
    calculator = DaycareSubsidyCalculator(app.session())

    services = Services.from_org(app.org)
    services.select('ganzer-tag-inkl-mitagessen', 'mo')

    calculation = calculator.calculate(
        daycare=calculator.daycare_by_title("Luftibus"),
        services=services,
        income=Decimal('20000'),
        wealth=Decimal('10000'),
        rebate=True,
    )

    base, gross, net, actual, monthly = calculation.blocks

    # note, these results slightly differ from the output, as the rounding
    # only happens when the numbers are rendered
    results = [(r.title, r.amount) for r in base.results]
    assert results == [
        ("Steuerbares Einkommen", Decimal('20000')),
        ("Vermögenszuschlag", Decimal('0')),
        ("Massgebendes Gesamteinkommen", Decimal('20000')),
        ("Abzüglich Minimaleinkommen", Decimal('20000')),
        ("Berechnungsgrundlage", Decimal('0')),
    ]

    results = [(r.title, r.amount) for r in gross.results]
    assert results == [
        ("Übertrag", Decimal('0')),
        ("Faktor", Decimal("0.001672727")),
        ("Einkommensabhängiger Elternbeitragsbestandteil", Decimal("0")),
        ("Mindestbeitrag Eltern", Decimal("15")),
        ("Elternbeitrag brutto", Decimal("15")),
    ]

    results = [(r.title, r.amount) for r in net.results]
    assert results == [
        ("Übertrag", Decimal('15')),
        ("Rabatt", Decimal('0.75')),
        ("Elternbeitrag netto", Decimal('15')),
    ]

    results = [(r.title, r.amount) for r in actual.results]
    assert results == [
        ("Übertrag", Decimal('15')),
        ("Zusatzbeitrag Eltern", Decimal('3')),
        ("Elternbeitrag pro Tag", Decimal('18')),
        ("Städtischer Beitrag pro Tag", Decimal('92'))
    ]

    results = [(r.title, r.amount) for r in monthly.results]
    assert results == [
        ("Wochentarif", Decimal('18')),
        ("Faktor", Decimal('4.25')),
        ("Elternbeitrag pro Monat", Decimal('76.50')),
        ("Städtischer Beitrag pro Monat", Decimal('391')),
    ]


def test_calculate_example_5(app):
    calculator = DaycareSubsidyCalculator(app.session())

    services = Services.from_org(app.org)
    services.select('ganzer-tag-inkl-mitagessen', 'mo')
    services.select('vor-oder-nachmittag-inkl-mitagessen', 'di')
    services.select('vor-oder-nachmittag-inkl-mitagessen', 'mi')

    calculation = calculator.calculate(
        daycare=calculator.daycare_by_title("Am Park"),
        services=services,
        income=Decimal('42000'),
        wealth=Decimal('15000'),
        rebate=False,
    )

    base, gross, net, actual, monthly = calculation.blocks

    # note, these results slightly differ from the output, as the rounding
    # only happens when the numbers are rendered
    results = [(r.title, r.amount) for r in base.results]
    assert results == [
        ("Steuerbares Einkommen", Decimal('42000')),
        ("Vermögenszuschlag", Decimal('0')),
        ("Massgebendes Gesamteinkommen", Decimal('42000')),
        ("Abzüglich Minimaleinkommen", Decimal('20000')),
        ("Berechnungsgrundlage", Decimal('22000')),
    ]

    results = [(r.title, r.amount) for r in gross.results]
    assert results == [
        ("Übertrag", Decimal('22000')),
        ("Faktor", Decimal("0.001672727")),
        ("Einkommensabhängiger Elternbeitragsbestandteil", Decimal("36.80")),
        ("Mindestbeitrag Eltern", Decimal("15")),
        ("Elternbeitrag brutto", Decimal("51.80")),
    ]

    results = [(r.title, r.amount) for r in net.results]
    assert results == [
        ("Übertrag", Decimal('51.80')),
        ("Rabatt", Decimal('0.0')),
        ("Elternbeitrag netto", Decimal('51.80')),
    ]

    results = [(r.title, r.amount) for r in actual.results]
    assert results == [
        ("Übertrag", Decimal('51.80')),
        ("Zusatzbeitrag Eltern", Decimal('13')),
        ("Elternbeitrag pro Tag", Decimal('64.80')),
        ("Städtischer Beitrag pro Tag", Decimal('55.20'))
    ]

    results = [(r.title, r.amount) for r in monthly.results]
    assert results == [
        ("Wochentarif", Decimal('162')),
        ("Faktor", Decimal('4.0833')),
        ("Elternbeitrag pro Monat", Decimal('661.50')),
        ("Städtischer Beitrag pro Monat", Decimal('563.50')),
    ]



import textwrap

from onegov.core.security import Secret
from onegov.directory import Directory, DirectoryCollection
from onegov.form import Form
from onegov.form.fields import HtmlField
from onegov.org.models import Organisation
from onegov.org.views.settings import handle_generic_settings
from onegov.winterthur import _
from onegov.winterthur.app import WinterthurApp
from onegov.winterthur.daycare import Services
from wtforms.fields import RadioField, TextAreaField
from wtforms.fields.html5 import DecimalField
from wtforms.validators import InputRequired, ValidationError
from yaml.error import YAMLError


class WinterthurDaycareSettingsForm(Form):

    max_income = DecimalField(
        label=_("Maximum taxable income"),
        fieldset=_("Variables"),
        places=0,
        validators=[InputRequired()])

    max_wealth = DecimalField(
        label=_("Maximum taxable wealth"),
        fieldset=_("Variables"),
        places=0,
        validators=[InputRequired()])

    min_income = DecimalField(
        label=_("Minimum income"),
        fieldset=_("Variables"),
        places=0,
        validators=[InputRequired()])

    min_rate = DecimalField(
        label=_("Minimum day-rate"),
        fieldset=_("Variables"),
        places=0,
        validators=[InputRequired()])

    max_rate = DecimalField(
        label=_("Maximum day-rate"),
        fieldset=_("Variables"),
        places=0,
        validators=[InputRequired()])

    max_subsidy = DecimalField(
        label=_("Maximum subsidy"),
        fieldset=_("Variables"),
        places=0,
        validators=[InputRequired()])

    wealth_premium = DecimalField(
        label=_("Wealth premium (%)"),
        fieldset=_("Variables"),
        places=2,
        validators=[InputRequired()])

    rebate = DecimalField(
        label=_("Rebate (%)"),
        fieldset=_("Variables"),
        places=2,
        validators=[InputRequired()])

    services = TextAreaField(
        label=_("Care"),
        fieldset=_("Variables"),
        validators=[InputRequired()],
        render_kw={'rows': 32, 'data-editor': 'yaml'})

    directory = RadioField(
        label=_("Directory"),
        fieldset=_("Institutions"),
        validators=[InputRequired()],
        choices=None)

    explanation = HtmlField(
        label=_("Explanation"),
        fieldset=_("Details"),
        render_kw={'rows': 32})

    def populate_obj(self, obj, *args, **kwargs):
        super().populate_obj(obj, *args, **kwargs)
        obj.meta['daycare_settings'] = {
            k: v for k, v in self.data.items() if k != 'csrf_token'
        }

    def process_obj(self, obj):
        super().process_obj(obj)
        for k, v in obj.meta.get('daycare_settings', {}).items():
            if hasattr(self, k):
                getattr(self, k).data = v

        if not self.services.data or not self.services.data.strip():
            self.services.data = textwrap.dedent("""
                # Beispiel:
                #
                # - titel: "Ganzer Tag inkl. Mitagessen"
                #   tage: "Montag, Dienstag, Mittwoch, Donnerstag, Freitag"
                #   prozent: 100.00
            """)

    def validate_services(self, field):
        try:
            tuple(Services.parse_definition(field.data))
        except (YAMLError, TypeError, KeyError):
            raise ValidationError(_("Invalid services configuration"))

    def directory_choices(self):
        dirs = DirectoryCollection(self.request.session, type='extended')

        def choice(directory):
            return (
                directory.id.hex,
                directory.title
            )

        for d in dirs.query().order_by(Directory.order):
            yield choice(d)

    def on_request(self):
        self.directory.choices = list(self.directory_choices())


@WinterthurApp.form(model=Organisation, name='daycare-settings',
                    template='form.pt', permission=Secret,
                    form=WinterthurDaycareSettingsForm,
                    setting=_("Daycare Calculator"),
                    icon='fa-calculator')
def custom_handle_settings(self, request, form):
    return handle_generic_settings(self, request, form, _("Daycare Settings"))



from onegov.winterthur.models.address import WinterthurAddress
from onegov.winterthur.models.mission_report import MissionReport
from onegov.winterthur.models.mission_report import MissionReportFile
from onegov.winterthur.models.mission_report import MissionReportVehicle
from onegov.winterthur.models.mission_report import MissionReportVehicleUse


__all__ = (
    'WinterthurAddress',
    'MissionReport',
    'MissionReportFile',
    'MissionReportVehicle',
    'MissionReportVehicleUse',
)



from datetime import date
from onegov.winterthur.app import WinterthurApp
from onegov.winterthur.collections import AddressCollection
from onegov.winterthur.collections import AddressSubsetCollection
from onegov.winterthur.collections import MissionReportCollection
from onegov.winterthur.collections import MissionReportFileCollection
from onegov.winterthur.collections import MissionReportVehicleCollection
from onegov.winterthur.daycare import DaycareSubsidyCalculator
from onegov.winterthur.models import MissionReport
from onegov.winterthur.models import MissionReportVehicle
from onegov.winterthur.roadwork import RoadworkCollection, Roadwork
from uuid import UUID


@WinterthurApp.path(
    model=AddressCollection,
    path='/streets')
def get_streets_directory(app):
    return AddressCollection(app.session())


@WinterthurApp.path(
    model=AddressSubsetCollection,
    path='/streets/{street}')
def get_street_subset(app, street):
    subset = AddressSubsetCollection(app.session(), street=street)
    return subset.exists() and subset or None


@WinterthurApp.path(
    model=RoadworkCollection,
    path='/roadwork')
def get_roadwork_collection(app, letter=None, query=None):
    return RoadworkCollection(app.roadwork_client, letter=letter, query=query)


@WinterthurApp.path(
    model=Roadwork,
    path='/roadwork/{id}',
    converters=dict(id=int))
def get_roadwork(app, id):
    return RoadworkCollection(app.roadwork_client).by_id(id)


@WinterthurApp.path(
    model=MissionReportCollection,
    path='/mission-reports')
def get_mission_reports(request, page=0, year=date.today().year):
    year = str(year).isdigit() and int(year) or date.today().year

    return MissionReportCollection(
        request.session, page=page,
        include_hidden=request.is_manager,
        year=year)


@WinterthurApp.path(
    model=MissionReportVehicleCollection,
    path='/mission-reports/vehicles')
def get_mission_report_vehicles(request):
    return MissionReportVehicleCollection(request.session)


@WinterthurApp.path(
    model=MissionReport,
    path='/mission-reports/report/{id}',
    converters=dict(id=UUID))
def get_mission_report(request, id):
    return get_mission_reports(request).by_id(id)


@WinterthurApp.path(
    model=MissionReportFileCollection,
    path='/mission-reports/report/{id}/images',
    converters=dict(id=UUID))
def get_mission_report_files(request, id):
    report = get_mission_report(request, id)

    if report:
        return MissionReportFileCollection(request.session, report)


@WinterthurApp.path(
    model=MissionReportVehicle,
    path='/mission-reports/vehicle/{id}',
    converters=dict(id=UUID))
def get_mission_report_vehicle(request, id):
    return get_mission_report_vehicles(request).by_id(id)


@WinterthurApp.path(
    model=DaycareSubsidyCalculator,
    path='/daycare-subsidy-calculator')
def get_daycare_subsidy_calculator(request):
    return DaycareSubsidyCalculator(request.session)



from cached_property import cached_property
from onegov.org.layout import DefaultLayout
from onegov.core.elements import Link, LinkGroup, Intercooler, Confirm, Block
from onegov.winterthur import _
from onegov.winterthur.collections import AddressCollection
from onegov.winterthur.collections import MissionReportCollection
from onegov.winterthur.collections import MissionReportFileCollection
from onegov.winterthur.collections import MissionReportVehicleCollection
from onegov.winterthur.models import MissionReport
from onegov.winterthur.models import MissionReportVehicle
from onegov.winterthur.roadwork import RoadworkCollection


class AddressLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Addresses"), '#'),
        ]

    @cached_property
    def editbar_links(self):
        if not self.request.is_manager:
            return

        return [
            Link(
                text=_("Update"),
                url=self.csrf_protected_url(
                    self.request.link(self.model, '+update')
                ),
                attrs={'class': 'sync'},
                traits=Intercooler(
                    request_method='POST',
                    redirect_after=self.request.url
                )
            )
        ]


class AddressSubsetLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Addresses"), self.request.class_link(AddressCollection)),
            Link(_(self.model.street), '#')
        ]


class RoadworkCollectionLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Roadworks"), '#'),
        ]


class RoadworkLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return [
            Link(_("Homepage"), self.homepage_url),
            Link(_("Roadworks"), self.request.class_link(RoadworkCollection)),
            Link(self.model.title, self.request.link(self.model))
        ]


class DaycareSubsidyCalculatorLayout(DefaultLayout):

    @cached_property
    def breadcrumbs(self):
        return (
            Link(_("Homepage"), self.homepage_url),
            Link(
                _("Daycare Subsidy Calculator"),
                self.request.link(self.model)
            )
        )


class MissionReportLayout(DefaultLayout):

    def __init__(self, model, request, *suffixes):
        self.suffixes = suffixes
        super().__init__(model, request)
        request.include('iframe-enhancements')

    def breadcrumbs_iter(self):
        yield Link(
            _("Homepage"),
            self.homepage_url)

        yield Link(
            _("Mission Reports"),
            self.request.class_link(MissionReportCollection))

        yield from self.suffixes

    @cached_property
    def breadcrumbs(self):
        return list(self.breadcrumbs_iter())

    @cached_property
    def image_upload_url(self):

        if not hasattr(self.model, 'report'):
            return super().image_upload_url

        url = self.request.link(
            MissionReportFileCollection(
                self.request.session, self.model.report), name='upload')

        return self.csrf_protected_url(url)

    @cached_property
    def image_upload_json_url(self):

        if not hasattr(self.model, 'report'):
            return super().image_upload_json_url

        url = self.request.link(
            MissionReportFileCollection(
                self.request.session, self.model.report), name='upload.json')

        return self.csrf_protected_url(url)

    @cached_property
    def editbar_links(self):

        # this is a bit different then usual, trying out some things as part
        # of this project - probably not a good idea to copy this elsewhere
        if not self.request.is_manager:
            return

        if self.suffixes and not getattr(self.suffixes[-1], 'editbar', True):
            return

        if isinstance(self.model, MissionReportCollection):
            return [
                Link(
                    _("Vehicles"), self.request.class_link(
                        MissionReportVehicleCollection
                    ), attrs={'class': 'vehicles'}
                ),
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Mission Report"),
                            url=self.request.link(
                                self.model,
                                name='+new'
                            ),
                            attrs={'class': 'new-report'}
                        )
                    ]
                ),
            ]

        if isinstance(self.model, MissionReportVehicleCollection):
            return [
                LinkGroup(
                    title=_("Add"),
                    links=[
                        Link(
                            text=_("Vehicle"),
                            url=self.request.link(
                                self.model,
                                name='+new'
                            ),
                            attrs={'class': 'new-vehicle'}
                        )
                    ]
                ),
            ]

        if isinstance(self.model, MissionReport):
            return [
                Link(
                    _("Images"),
                    self.request.link(self.model, name='images'),
                    attrs={'class': 'upload-images'}
                ),
                Link(
                    _("Edit"),
                    self.request.link(self.model, name='edit'),
                    attrs={'class': 'edit-link'}
                ),
                Link(
                    _("Delete"),
                    self.csrf_protected_url(
                        self.request.link(self.model)
                    ),
                    attrs={'class': 'delete-link'},
                    traits=(
                        Confirm(
                            _(
                                "Do you really want to delete "
                                "this mission report?"
                            ),
                            _("This cannot be undone."),
                            _("Delete mission report"),
                            _("Cancel")
                        ),
                        Intercooler(
                            request_method='DELETE',
                            redirect_after=self.request.class_link(
                                MissionReportCollection
                            )
                        )
                    )
                )
            ]

        if isinstance(self.model, MissionReportVehicle):
            if self.model.uses:
                return [
                    Link(
                        _("Delete"),
                        '#',
                        attrs={'class': 'delete-link'},
                        traits=(
                            Block(
                                _("This vehicle can't be deleted."),
                                _(
                                    "There are mission reports associated "
                                    "with this vehicle."
                                ),
                                _("Cancel")
                            ),
                        )
                    )
                ]
            else:
                return [
                    Link(
                        _("Delete"),
                        self.csrf_protected_url(
                            self.request.link(self.model)
                        ),
                        attrs={'class': 'delete-link'},
                        traits=(
                            Confirm(
                                _(
                                    "Do you really want to delete "
                                    "this vehicle?"
                                ),
                                _("This cannot be undone."),
                                _("Delete vehicle"),
                                _("Cancel")
                            ),
                            Intercooler(
                                request_method='DELETE',
                                redirect_after=self.request.class_link(
                                    MissionReportVehicleCollection
                                )
                            )
                        )
                    )
                ]



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""

from onegov.core.upgrade import upgrade_task
from onegov.org.models import Organisation


@upgrade_task('Change the default geo provider', always_run=True)
def change_default_geo_provider(context):

    org = context.session.query(Organisation).first()

    if org is None:
        return False

    if "Strassenverzeichnis" not in org.meta['homepage_structure']:
        return False

    org.meta['geo_provider'] = 'geo-vermessungsamt-winterthur'



from onegov.winterthur.forms.mission_report import MissionReportForm
from onegov.winterthur.forms.mission_report import MissionReportVehicleForm

__all__ = ('MissionReportForm', 'MissionReportVehicleForm')



from onegov.core.security import Public
from onegov.winterthur import WinterthurApp, _
from onegov.winterthur.layout import RoadworkLayout, RoadworkCollectionLayout
from onegov.winterthur.roadwork import Roadwork, RoadworkCollection


@WinterthurApp.html(
    model=RoadworkCollection,
    permission=Public,
    template='roadworks.pt'
)
def view_roadwork_collection(self, request):

    return {
        'layout': RoadworkCollectionLayout(self, request),
        'title': _("Roadworks"),
        'model': self
    }


@WinterthurApp.html(
    model=Roadwork,
    permission=Public,
    template='roadwork.pt'
)
def view_roadwork(self, request):

    return {
        'layout': RoadworkLayout(self, request),
        'title': self.title,
        'model': self,
        'back': request.class_link(RoadworkCollection)
    }



import chameleon
import textwrap
import yaml

from babel.numbers import format_decimal
from cached_property import cached_property
from collections import defaultdict
from collections import OrderedDict
from decimal import Decimal, localcontext
from onegov.core.utils import Bunch
from onegov.core.utils import normalize_for_url
from onegov.directory import DirectoryCollection
from onegov.directory import DirectoryEntryCollection
from onegov.form import Form
from onegov.org.models import Organisation
from onegov.winterthur import _
from ordered_set import OrderedSet
from wtforms.fields import Field, BooleanField, SelectField
from wtforms.fields.html5 import DecimalField
from wtforms.validators import NumberRange, InputRequired, ValidationError
from wtforms.widgets.core import HTMLString


SERVICE_DAYS = {
    'mo': 0,
    'di': 1,
    'mi': 2,
    'do': 3,
    'fr': 4,
    'sa': 5,
    'so': 6,
}

SERVICE_DAYS_LABELS = {
    0: _("Monday"),
    1: _("Tuesday"),
    2: _("Wednesday"),
    3: _("Thursday"),
    4: _("Friday"),
    5: _("Saturday"),
    6: _("Sunday"),
}

# http://babel.pocoo.org/en/latest/numbers.html#pattern-syntax
FORMAT = '#,##0.00########'


def round_to(n, precision):
    assert isinstance(precision, str)

    precision = Decimal(precision)
    correction = Decimal('0.5') if n >= 0 else Decimal('-0.5')

    return int(n / precision + correction) * precision


def format_precise(amount):
    if not amount:
        return '0.00'

    with localcontext() as ctx:
        ctx.prec = 28

        return format_decimal(amount, format=FORMAT, locale='de_CH')


def format_1_cent(amount):
    return format_precise(round_to(amount, '0.01'))


def format_5_cents(amount):
    return format_precise(round_to(amount, '0.05'))


class Daycare(object):

    def __init__(self, id, title, rate, weeks):
        self.id = id
        self.title = title
        self.rate = Decimal(rate)
        self.weeks = weeks

    @property
    def factor(self):
        return Decimal(self.weeks) / Decimal('12')


class Services(object):

    def __init__(self, definition):
        if definition:
            self.available = OrderedDict(self.parse_definition(definition))
        else:
            self.available = OrderedDict()

        self.selected = defaultdict(set)

    @classmethod
    def from_org(cls, org):
        if 'daycare_settings' not in org.meta:
            return cls(None)

        if 'services' not in org.meta['daycare_settings']:
            return cls(None)

        return cls(org.meta['daycare_settings']['services'])

    @classmethod
    def from_session(cls, session):
        return cls.from_org(session.query(Organisation).one())

    @staticmethod
    def parse_definition(definition):
        for service in yaml.safe_load(definition):
            service_id = normalize_for_url(service['titel'])
            days = (d.strip() for d in service['tage'].split(','))

            yield service_id, Bunch(
                id=service_id,
                title=service['titel'],
                percentage=Decimal(service['prozent']),
                days=OrderedSet(SERVICE_DAYS[d.lower()[:2]] for d in days),
            )

    def select(self, service_id, day):
        self.selected[service_id].add(day)

    def deselect(self, service_id, day):
        self.selected[service_id].remove(day)

    def is_selected(self, service_id, day):
        if service_id not in self.selected:
            return False

        return day in self.selected[service_id]

    @property
    def total(self):
        """ Returns the total percentage of used services. """
        return sum(
            self.available[s].percentage * len(self.selected[s])
            for s in self.selected
        )


class Result(object):

    def __init__(self, title, amount=None, note=None, operation=None,
                 important=False, currency='CHF', output_format=None):

        self.title = title
        self.amount = amount
        self.note = textwrap.dedent(note or '').strip(' \n')
        self.operation = operation
        self.important = important
        self.currency = currency
        self.output_format = output_format or format_1_cent

    def __bool__(self):
        return bool(self.amount)

    @property
    def readable_amount(self):
        return self.output_format(self.amount)


class Block(object):

    def __init__(self, id, title):
        self.id = id
        self.title = title
        self.results = []
        self.total = Decimal(0)

    def op(self, title, amount=None, note=None, operation=None,
           important=False, currency='CHF', output_format=None,
           total_places=2, amount_places=2):

        if amount == 0:
            amount = Decimal('0')

        def limit_total(total):
            return total.quantize(Decimal(f'0.{"0" * (total_places - 1)}1'))

        def limit_amount(amount):
            return amount.quantize(Decimal(f'0.{"0" * (amount_places - 1)}1'))

        if operation is None:
            assert amount is not None
            self.total = amount

        elif operation == '+':
            assert amount is not None
            self.total += amount

        elif operation == '=':
            amount = self.total if amount is None else amount
            self.total = max(amount, Decimal('0'))

        elif operation == '-':
            assert amount is not None
            self.total -= amount

        elif operation in ('*', 'x', '×', '⋅'):
            assert amount is not None
            self.total *= amount

        elif operation in ('/', '÷'):
            assert amount is not None
            self.total /= amount

        # limit the amount and the total after the operation, not before
        self.total = limit_total(self.total)
        amount = limit_amount(amount)

        self.results.append(Result(
            title=title,
            amount=amount,
            note=note,
            operation=operation,
            important=important,
            currency=currency,
            output_format=output_format,
        ))

        return self.total


class DirectoryDaycareAdapter(object):

    def __init__(self, directory):
        self.directory = directory

    @cached_property
    def fieldmap(self):
        fieldmap = {
            'daycare_rate': None,
            'daycare_weeks': None,
            'daycare_url': None,
        }

        for field in self.directory.basic_fields:

            if 'tarif' in field.label.lower():
                fieldmap['daycare_rate'] = field.id
                continue

            if 'woche' in field.label.lower():
                fieldmap['daycare_weeks'] = field.id
                continue

            if 'web' in field.label.lower():
                fieldmap['daycare_url'] = field.id
                continue

        return fieldmap

    def as_daycare(self, entry):
        return Daycare(
            id=entry.id,
            title=entry.title,
            rate=entry.values[self.fieldmap['daycare_rate']],
            weeks=entry.values[self.fieldmap['daycare_weeks']],
        )


class Settings(object):

    def __init__(self, organisation):
        settings = organisation.meta.get('daycare_settings', {})

        for key, value in settings.items():
            setattr(self, key, value)

    def is_valid(self):
        keys = (
            'directory',
            'max_income',
            'max_rate',
            'max_subsidy',
            'max_wealth',
            'min_income',
            'min_rate',
            'rebate',
            'services',
            'wealth_premium',
        )

        for key in keys:
            if not hasattr(self, key):
                return False

        return True

    def factor(self, daycare):
        min_day_rate = daycare.rate - self.min_rate
        min_day_rate = min(min_day_rate, self.max_subsidy)

        factor = min_day_rate / (self.max_income - self.min_income)
        factor = factor.quantize(Decimal('0.000000001'))

        return factor


class DaycareSubsidyCalculator(object):

    def __init__(self, session):
        self.session = session

    @cached_property
    def organisation(self):
        return self.session.query(Organisation).one()

    @cached_property
    def settings(self):
        return Settings(self.organisation)

    @cached_property
    def directory(self):
        return DirectoryCollection(self.session).by_id(self.settings.directory)

    @cached_property
    def daycares(self):
        adapter = DirectoryDaycareAdapter(self.directory)

        items = DirectoryEntryCollection(self.directory).query()
        items = (i for i in items if not i.meta.get('is_hidden_from_public'))
        items = {i.id.hex: adapter.as_daycare(i) for i in items}

        return items

    def daycare_by_title(self, title):
        return next(d for d in self.daycares.values() if d.title == title)

    def calculate(self, *args, **kwargs):
        return self.calculate_precisely(*args, **kwargs)

    def calculate_precisely(self, daycare, services, income, wealth, rebate):
        """ Creates a detailed calculation of the subsidy paid by Winterthur.

        The reslt is a list of tables with explanations.

        :param daycare:
            The selected daycare (a :class:`Daycare` instance).

        :param services:
            Services used (a :class:`Services` instance)

        :param income:
            The income as a decimal.

        :param wealth:
            The wealth as decimal.

        :param rebate:
            True if a rebate is applied

        Note, due to the specific nature of the content here, which is probably
        not going to be translated, we use German. For consistency we want to
        limit this, but with Winterthur these kinds of things crop up as the
        wording is quite specific and adding translations would just make
        this a lot harder.

        """

        cfg = self.settings
        fmt = format_precise

        # Base Rate
        # ---------
        base = Block('base', "Berechnungsgrundlage für die Elternbeiträge")

        base.op(
            title="Steuerbares Einkommen",
            amount=income,
            note="""
                Steuerbares Einkommen gemäss letzter Veranlagung.
            """)

        base.op(
            title="Vermögenszuschlag",
            amount=max(
                (wealth - cfg.max_wealth)
                * cfg.wealth_premium
                / Decimal('100'),
                0),
            operation="+",
            note=f"""
                Der Vermögenszuschlag beträgt {fmt(cfg.wealth_premium)}% des
                Vermögens, für das tatsächlich Steuern anfallen
                (ab {fmt(cfg.max_wealth)} CHF).
            """)

        base.op(
            title="Massgebendes Gesamteinkommen",
            operation="=")

        base.op(
            title="Abzüglich Minimaleinkommen",
            operation="-",
            amount=cfg.min_income)

        base.op(
            title="Berechnungsgrundlage",
            operation="=")

        # Gross Contribution
        # ------------------
        gross = Block('gross', "Berechnung des Brutto-Elternbeitrags")

        gross.op(
            title="Übertrag",
            amount=base.total)

        gross.op(
            title="Faktor",
            amount=cfg.factor(daycare),
            currency=None,
            operation="×",
            note="""
                Ihr Elternbeitrag wird aufgrund eines Faktors berechnet
                (Kita-Reglement Art. 20 Abs 3).
            """,
            output_format=format_precise,
            amount_places=10)

        gross.op(
            title="Einkommensabhängiger Elternbeitragsbestandteil",
            operation="=")

        gross.op(
            title="Mindestbeitrag Eltern",
            amount=cfg.min_rate,
            operation="+")

        gross.op(
            title="Elternbeitrag brutto",
            operation="=",
            amount=min(gross.total, daycare.rate))

        # Rebate
        # ------
        rebate = gross.total * cfg.rebate / 100 if rebate else 0

        net = Block('net', "Berechnung des Rabatts")

        net.op(
            title="Übertrag",
            amount=gross.total)

        net.op(
            title="Rabatt",
            amount=rebate,
            operation="-",
            note=f"""
                Bei einem Betreuungsumfang von insgesamt mehr als 2 ganzen
                Tagen pro Woche gilt ein Rabatt von {cfg.rebate}%.
            """)

        net.op(
            title="Elternbeitrag netto",
            operation="=",
            amount=max(cfg.min_rate, gross.total - rebate))

        # Actual contribution
        # -------------------
        actual = Block('actual', (
            "Berechnung des Elternbeitrags und des "
            "städtischen Beitrags pro Tag"
        ))

        actual.op(
            title="Übertrag",
            amount=net.total)

        actual.op(
            title="Zusatzbeitrag Eltern",
            amount=max(daycare.rate - cfg.max_rate, 0),
            operation="+",
            note=f"""
                Zusatzbeitrag für Kitas, deren Tagestarif über
                {cfg.max_rate} CHF liegt.
            """)

        parent_share_per_day = actual.op(
            title="Elternbeitrag pro Tag",
            operation="=",
            note="""
                Ihr Beitrag pro Tag (100%) und Kind.
            """,
            important=True)

        city_share_per_day = actual.op(
            title="Städtischer Beitrag pro Tag",
            amount=max(daycare.rate - parent_share_per_day, Decimal('0.00')),
            important=True,
            note="""
                Städtischer Beitrag für Ihr Kind pro Tag.
            """)

        # Monthly contribution
        # --------------------
        monthly = Block(
            'monthly', (
                "Berechnung des Elternbeitrags und des städtischen "
                "Beitrags pro Monat"
            )
        )

        monthly.op(
            title="Wochentarif",
            amount=parent_share_per_day * services.total / 100,
            note="""
                Wochentarif: Elternbeiträge der gewählten Betreuungstage.
            """)

        monthly.op(
            title="Faktor",
            amount=daycare.factor,
            currency=None,
            operation="×",
            note="""
                Faktor für jährliche Öffnungswochen Ihrer Kita.
            """,
            output_format=format_precise,
            amount_places=4)

        parent_share_per_month = monthly.op(
            title="Elternbeitrag pro Monat",
            operation="=",
            important=True,
            output_format=format_5_cents)

        city_share_per_month = monthly.op(
            title="Städtischer Beitrag pro Monat",
            amount=city_share_per_day * services.total / 100 * daycare.factor,
            important=True,
            note="""
                Städtischer Beitrag für Ihr Kind pro Monat.
            """,
            output_format=format_5_cents)

        # Services table
        # --------------
        def services_table():
            total = Decimal(0)
            total_percentage = Decimal(0)

            for day in SERVICE_DAYS.values():
                for service_id in services.selected:
                    if day in services.selected[service_id]:
                        service = services.available[service_id]
                        cost = parent_share_per_day * service.percentage / 100

                        total += cost
                        total_percentage += service.percentage

                        label = SERVICE_DAYS_LABELS[day]
                        yield (label, service.title, format_5_cents(cost))

            yield (_("Total"), None, format_5_cents(total))

        total = round_to(parent_share_per_month, '0.05')\
            + round_to(city_share_per_month, '0.05')

        return Bunch(
            blocks=(base, gross, net, actual, monthly),
            parent_share_per_month=format_5_cents(parent_share_per_month),
            city_share_per_month=format_5_cents(city_share_per_month),
            total_per_month=format_5_cents(total),
            agenda=tuple(services_table()),
        )


class DaycareServicesWidget(object):

    template = chameleon.PageTemplate("""
        <table class="daycare-services">
            <thead>
                <tr>
                    <th></th>
                    <th tal:repeat="service this.services.available.values()">
                        <div class="daycare-services-title">
                            ${service.title}
                        </div>
                        <div class="daycare-services-percentage">
                            ${service.percentage}%
                        </div>
                    </th>
                </tr>
            </thead>
            <tbody>
                <tr tal:repeat="day this.days">
                    <th>
                        <strong class="show-for-small-only">
                            ${this.day_label(day)[:2]}
                        </strong>
                        <strong class="show-for-medium-up">
                            ${this.day_label(day)}
                        </strong>
                    </th>
                    <td tal:repeat="svc this.services.available.values()">
                        <label>
                            <input
                                type="checkbox"

                                id="${svc.id}-${day}"
                                name="${this.field.name}"
                                value="${svc.id}-${day}"

                                tal:attributes="
                                    checked this.is_selected(svc, day)
                                "
                            />
                        </label>
                    </td>
                </tr>
            </tbody>
        </table
    """)

    def __call__(self, field, **kwargs):
        self.field = field
        self.services = field.services

        return HTMLString(self.template.render(this=self))

    def is_selected(self, service, day):
        return self.services.is_selected(service.id, day)

    def day_label(self, day):
        return self.field.meta.request.translate(SERVICE_DAYS_LABELS[day])

    @property
    def days(self):
        days = OrderedSet()

        for service in self.services.available.values():
            for day in service.days:
                days.add(day)

        return days


class DaycareServicesField(Field):

    widget = DaycareServicesWidget()

    @cached_property
    def services(self):
        return Services.from_session(self.meta.request.session)

    def process_formdata(self, valuelist):
        for value in valuelist:
            service_id, day = value.rsplit('-', maxsplit=1)
            self.services.select(service_id, int(day))

    def pre_validate(self, form):
        for day in SERVICE_DAYS.values():
            days = sum(
                1 for id in self.services.available
                if self.services.is_selected(id, day)
            )

            if days > 1:
                raise ValidationError(_("Each day may only be selected once."))


class DaycareSubsidyCalculatorForm(Form):

    daycare = SelectField(
        label=_("Daycare"),
        validators=(InputRequired(), ),
        choices=(), )

    services = DaycareServicesField(
        label=_("Care"),
        validators=(InputRequired(), ))

    income = DecimalField(
        label=_("Taxable income"),
        validators=(InputRequired(), NumberRange(min=0)))

    wealth = DecimalField(
        label=_("Taxable wealth"),
        validators=(InputRequired(), NumberRange(min=0)))

    rebate = BooleanField(
        label=_("Rebate"),
        description=_(
            "Does at least one child in your household attend the same "
            "daycare for more than two whole days a week?"
        ))

    def on_request(self):
        self.daycare.choices = tuple(self.daycare_choices)

    @property
    def daycare_choices(self):

        def choice(daycare):
            label = _((
                "${title} / day rate CHF ${rate} / "
                "${weeks} weeks open per year"
            ), mapping={
                'title': daycare.title,
                'rate': daycare.rate,
                'weeks': daycare.weeks
            })

            return (daycare.id.hex, self.request.translate(label))

        for daycare in self.model.daycares.values():
            yield choice(daycare)

    @property
    def selected_daycare(self):
        for daycare in self.model.daycares.values():
            if daycare.id.hex == self.daycare.data:
                return daycare
